# Will not test qr generator code as it is effectively tested
# where it is used in the sorter tests.
import pathlib
from qrImageIndexer import qr_generator
from os import path
from PIL import ImageChops


def build_demo_tsv(file_path : pathlib.Path):
    content = '''Level1-1
Level1-2
\tLevel2-1
\tLevel2-2
\t\tLevel3'''
    with open(file_path, 'w') as f:
        f.write(content)

def build_demo_tsv_four_spaces(file_path : pathlib.Path):
    content = '''Level1-1
Level1-2
    Level2-1
    Level2-2
        Level3'''
    with open(file_path, 'w') as f:
        f.write(content)


def demo_tsv_expected_data():
    data = []
    data.append(['Level1-1'])
    data.append(['Level1-2'])
    data.append(['', 'Level2-1'])
    data.append(['', 'Level2-2'])
    data.append(['', '', 'Level3'])
    return data

def demo_tsv_drop_index():
    data = []
    data.append(['Level1-1'])
    data.append(['', 'Level2-1'])
    data.append(['', 'Level2-2'])
    data.append(['', '', 'Level3'])
    data.append(['Level1-2'])
    return data

def demo_data_struct_drop_indexinclude_headers():
    test_struct = {}
    test_struct['Level1-1'] = ({}, 'Level1-1' + path.sep)
    test_struct['Level1-1'][0]['Level2-1'] = ({}, 'Level1-1' + path.sep + 'Level2-1' + path.sep)
    test_struct['Level1-1'][0]['Level2-2'] = ({}, 'Level1-1' + path.sep + 'Level2-2' + path.sep)
    test_struct['Level1-1'][0]['Level2-2'][0]['Level3'] = ({}, 'Level1-1' + path.sep + 'Level2-2' + path.sep + 'Level3' + path.sep)
    test_struct['Level1-2'] = ({}, 'Level1-2' + path.sep)
    return test_struct


def demo_data_blank_line_end():
    data = []
    data.append(['Level1-1'])
    data.append(['Level1-2'])
    data.append(['', 'Level2-1'])
    data.append(['', 'Level2-2'])
    data.append(['', '', 'Level3'])
    data.append([])
    return data

def demo_data_blank_line_mid():
    data = []
    data.append(['Level1-1'])
    data.append([])
    data.append(['Level1-2'])
    data.append(['', 'Level2-1'])
    data.append(['', 'Level2-2'])
    data.append(['', '', 'Level3'])
    return data

def demo_data_struct_include_headers():
    test_struct = {}
    test_struct['Level1-1'] = ({}, 'Level1-1' + path.sep)
    test_struct['Level1-2'] = ({}, 'Level1-2' + path.sep)
    test_struct['Level1-2'][0]['Level2-1'] = ({}, 'Level1-2' + path.sep + 'Level2-1' + path.sep)
    test_struct['Level1-2'][0]['Level2-2'] = ({}, 'Level1-2' + path.sep + 'Level2-2' + path.sep)
    test_struct['Level1-2'][0]['Level2-2'][0]['Level3'] = ({}, 'Level1-2' + path.sep + 'Level2-2' + path.sep + 'Level3' + path.sep)
    return test_struct

def demo_data_struct_include_headers_with_QRHeader():
    test_struct = {}
    test_struct['Level1-1'] = ({}, r'{image}Level1-1' + path.sep)
    test_struct['Level1-2'] = ({}, r'{image}Level1-2' + path.sep)
    test_struct['Level1-2'][0]['Level2-1'] = ({}, r'{image}Level1-2' + path.sep + 'Level2-1' + path.sep)
    test_struct['Level1-2'][0]['Level2-2'] = ({}, r'{image}Level1-2' + path.sep + 'Level2-2' + path.sep)
    test_struct['Level1-2'][0]['Level2-2'][0]['Level3'] = ({}, r'{image}Level1-2' + path.sep + 'Level2-2' + path.sep + 'Level3' + path.sep)
    return test_struct

def demo_data_struct_no_headers():
    test_struct = {}
    test_struct['Level1-1'] = ({}, 'Level1-1' + path.sep)
    test_struct['Level1-2'] = ({}, None)
    test_struct['Level1-2'][0]['Level2-1'] = ({}, 'Level1-2' + path.sep + 'Level2-1' + path.sep)
    test_struct['Level1-2'][0]['Level2-2'] = ({}, None)
    test_struct['Level1-2'][0]['Level2-2'][0]['Level3'] = ({}, 'Level1-2' + path.sep + 'Level2-2' + path.sep + 'Level3' + path.sep)
    return test_struct

def demo_data_struct_no_headers_images():
    test_struct = {}
    test_struct['Level1-1'] = ({}, qr_generator.build_qr('Level1-1' + path.sep))
    test_struct['Level1-2'] = ({}, None)
    test_struct['Level1-2'][0]['Level2-1'] = ({}, qr_generator.build_qr('Level1-2' + path.sep + 'Level2-1' + path.sep))
    test_struct['Level1-2'][0]['Level2-2'] = ({}, None)
    test_struct['Level1-2'][0]['Level2-2'][0]['Level3'] = ({}, qr_generator.build_qr('Level1-2' + path.sep + 'Level2-2' + path.sep + 'Level3' + path.sep))
    return test_struct

def demo_load_lines_data():
    return ["Test Line 1", "\tTest Line 2", "Test Line 3"]

def expected_load_lines_data():
    return [
        ["Test Line 1"],
        ["", "Test Line 2"],
        ["Test Line 3"],
    ]

def test_build_indent_string_print():
    assert '' == qr_generator.build_indent_string_print(0)
    assert '|___' == qr_generator.build_indent_string_print(1)
    for i in range(2, 10):
        assert '\t' * (i-1) + '|___' == qr_generator.build_indent_string_print(i)

def test_print_struct_outline(capsys):


    expected_out = '''Level1-1
Level1-2
|___Level2-1
|___Level2-2
\t|___Level3
'''

    qr_generator.print_struct_outline(demo_data_struct_include_headers())

    captured = capsys.readouterr()
    assert expected_out == captured.out

def test_load_lines():
    lines = demo_load_lines_data()
    result = qr_generator.load_lines(lines)
    assert result == expected_load_lines_data()
    

def test_load_text_file(tmp_path : pathlib.Path):
    demo_file = tmp_path / 'file.txt'
    build_demo_tsv(demo_file)
    loaded_data = qr_generator.load_text_file(demo_file.as_posix())
    assert demo_tsv_expected_data() == loaded_data

def test_load_text_file_spaces(tmp_path : pathlib.Path):
    demo_file = tmp_path / 'file.txt'
    build_demo_tsv_four_spaces(demo_file)
    loaded_data = qr_generator.load_text_file(demo_file.as_posix())
    assert demo_tsv_expected_data() == loaded_data



def test_count_leading_indent():
    data = ['non_empty_string']
    for i in range(10):
        assert i == qr_generator.count_leading_indent(data)
        data.insert(0, '')

def test_unpack_data_include_headers():
    result_struct = {}
    qr_generator.unpack_data_recurse(demo_tsv_expected_data(), True, result_struct)

    assert demo_data_struct_include_headers() == result_struct

def test_unpack_data_include_headers_recurse_equivalence():
    result_struct_1 = {}
    qr_generator.unpack_data_recurse(demo_tsv_expected_data(), True, result_struct_1)
    result_struct_2 = qr_generator.unpack_data(demo_tsv_expected_data(), True)

    assert result_struct_1 == result_struct_2

def test_unpack_data_blank_line_end():
    result_struct = {}
    qr_generator.unpack_data_recurse(demo_data_blank_line_end(), True, result_struct)

    assert demo_data_struct_include_headers() == result_struct

def test_unpack_data_blank_line_end_recurse_equivalence():
    result_struct_1 = {}
    qr_generator.unpack_data_recurse(demo_data_blank_line_end(), True, result_struct_1)
    result_struct_2 = qr_generator.unpack_data(demo_data_blank_line_end(), True)

    assert result_struct_1 == result_struct_2

def test_unpack_data_blank_line_mid():
    result_struct = {}
    qr_generator.unpack_data_recurse(demo_data_blank_line_mid(), True, result_struct)

    assert demo_data_struct_include_headers() == result_struct

def test_unpack_data_no_headers():
    result_struct = {}
    qr_generator.unpack_data_recurse(demo_tsv_expected_data(), False, result_struct)

    assert demo_data_struct_no_headers() == result_struct

def test_unpack_data_no_headers_recurse_equivalence():
    result_struct_1 = {}
    qr_generator.unpack_data_recurse(demo_tsv_expected_data(), False, result_struct_1)
    result_struct_2 = qr_generator.unpack_data(demo_tsv_expected_data(), False)

    assert result_struct_1 == result_struct_2

def test_unpack_data_include_headers_QRHeader():
    result_struct = {}
    qr_generator.unpack_data_recurse(demo_tsv_expected_data(), True, result_struct, r'{image}')

    assert demo_data_struct_include_headers_with_QRHeader() == result_struct

def test_unpack_data_include_headers_QRHeader_recurse_equivalence():
    result_struct_1 = {}
    qr_generator.unpack_data_recurse(demo_tsv_expected_data(), True, result_struct_1, r'{image}')
    result_struct_2 = qr_generator.unpack_data(demo_tsv_expected_data(), True, r'{image}')

    assert result_struct_1 == result_struct_2

def test_structure_qr_builder():
    expected_struct = demo_data_struct_no_headers_images()
    generated_struct = qr_generator.generate_qr_code_structure(demo_data_struct_no_headers())
    assert not ImageChops.difference(expected_struct['Level1-1'][1], generated_struct['Level1-1'][1]).getbbox()
    assert not ImageChops.difference(expected_struct['Level1-2'][0]['Level2-1'][1], generated_struct['Level1-2'][0]['Level2-1'][1]).getbbox()
    assert not ImageChops.difference(expected_struct['Level1-2'][0]['Level2-2'][0]['Level3'][1], generated_struct['Level1-2'][0]['Level2-2'][0]['Level3'][1]).getbbox()
    assert expected_struct['Level1-2'][0]['Level2-2'][1] == generated_struct['Level1-2'][0]['Level2-2'][1]

def test_unpack_data_drop_leve():
    result_struct = qr_generator.unpack_data(demo_tsv_drop_index(), True)

    assert result_struct == demo_data_struct_drop_indexinclude_headers()


from doctest import OutputChecker
from io import TextIOWrapper
from turtle import fillcolor
import qrcode
from PIL import Image
from typing import Dict, List, Tuple
from os import path
from csv import reader

def build_qr(data : str) -> Image.Image:
    """
    Build a QR code with selected settings. Will use maximum error correction;
    this should reduce risk of errors when taking photos which may be blurry or
    have other issues.
    """
    qr = qrcode.QRCode(version=None,
                        error_correction=qrcode.constants.ERROR_CORRECT_H,
                        box_size=10,
                        border = 4)

    qr.add_data(data)
    qr.make(fit=True)
    return qr.make_image(fill_color='black', back_color='white')

def load_text_file(path : str) -> List[List[str]]:
    """
        Builds a data structure from a text file

        Parameters:
            path: path to text file as a string
            qr_headers: boolean indicating whether headings in text file are to have QRs generated
            string_header: string to be set at the start of QR code data to help weed out other QR codes

        Returns:
            List of list for table-like file (e.g. CSV)
    """
    output_data_structure = []
    with open(path, 'r') as f:
        output_data_structure = load_lines(f)

    return output_data_structure

def load_lines(input_lines) -> List[List[str]]:
    lines = []
    output_data_structure = []
    for row in input_lines:
        lines.append(row.replace('    ', '\t'))
    csv = reader(lines , delimiter='\t')
    for row in csv:
        output_data_structure.append(row)
    return output_data_structure

def generate_qr_code_structure(data_structure : Dict[str, Tuple[Dict, str]]) -> Dict[str, Tuple[Dict, Image.Image]]:
    '''
        Converts a dictionary containing the QR code string into a dictionary with an Image in place of the string

        Parameters:
            data_structure: Data structure in format of dictionary with tuple containing a like dictionary and a string
            where the string is the data to be embedded in the QR code. If no QR code required string to be None

        Returns:
            Similar structure to input but with string replaced by an PIL image
    '''
    result = {}
    for key in data_structure:
        image = None
        if data_structure[key][1]:
            image = build_qr(data_structure[key][1])
        sub_struct = generate_qr_code_structure(data_structure[key][0])
        result[key] = (sub_struct, image)
    return result

def unpack_data(data : List[List[str]], gen_qr_headings : bool, string_header : str = '') -> Dict[str, Tuple[Dict, str]]:
    """
    Function to unpack tabulated text where items are grouped by tab depth. Calls the recursive function to remove the
    complexity from view of potential users.
    
    Inputs:
        data - List of string representing data
        gen_qr_headings - Boolean indicating if QR codes are to be built for headings or just for final elements
        string_header - String to include as header for all QR code values to help distinguish from general QR codes


    Returns:
        index - Data structure in dictionary as required by the remainder of the tool

    """
    _, data_structure = unpack_data_recurse(data, gen_qr_headings, string_header=string_header)
    return data_structure

def unpack_data_recurse(data : List[List[str]], gen_qr_headings : bool, data_structure : Dict[str, Tuple[Dict, str]] = None, string_header : str = '',
    index : int = 0, previous_levels : str = '') -> Tuple[int, Dict[str, Tuple[Dict, str]]]:
    """
    Function to unpack a tabulated text where items are grouped by tab depth. Called recursively for each loweer level of
    the data structure.

    Inputs:
        data - List of string representing data
        gen_qr_headings - Boolean indicating if QR codes are to be built for headings or just for final elements
        data_structure - List where output data is to be stored. This will be altered by the function. Output dictionary will contain
            key of str and string for image in a touple of (Dict, str) where the Dict is the data structure for subsequent levels of the file and the
            str is the content for the QR code where relevent
        string_header - String to include as header for all QR code values to help distinguish from general QR codes
        index - Index to start at. Defaults to zero but non-zero values used for recursion
        previous_levels - String representing the folder paths represented by previous levels of the data structure. This is embedded in the QR
            code for sorting


    Returns below in a Tuple:
        index - integer representing current index of the process
        data_struct - Data structure in dictionary as required by the remainder of the tool

        
    """
    if data_structure is None:
        data_structure = {}
    #Determine target indent from first entry. All subsequent should be the same.
    target_indent = count_leading_indent(data[index])
    while index < len(data):
        # Check if the line is blank. If so we will increment to the next index and continue
        if data[index] == []:
            index += 1
            continue
        # Check if this index is less indented than the target. If so we need to return a level.
        if count_leading_indent(data[index]) < target_indent:
            return index, data_structure

        # Get some data for this iteration
        raw_line = data[index][target_indent]
        next_data_struct = {}
        next_level_str = previous_levels + raw_line + path.sep

        # Get the difference between the current indent and next indent levels. If no further lines, then indent diff of zero
        if index == (len(data) - 1):
            indent_diff = 0
        else:
            indent_diff = target_indent - count_leading_indent(data[index + 1])

        # If for whatever reason the entry already exists then get the existing dictionary to pass to the next level in
        if raw_line in data_structure:
            next_data_struct = data_structure[raw_line][0]
        else:
            image_str = None
            if gen_qr_headings or indent_diff >= 0:
                image_str = string_header + next_level_str
            data_structure[raw_line] = (next_data_struct, image_str)

        # Check if the next line is further indented or if it is less indented. If more indented then recursively call function.
        # Otherwise increment
        if indent_diff < 0:
            index, _ = unpack_data_recurse(data, gen_qr_headings, next_data_struct, string_header, index + 1, next_level_str)
        else:
            index += 1

        
    return index, data_structure

def count_leading_indent(line : List[str]) -> int:
    """
        Counts leacding tabs in a passed string. Will convert sets of 4 spaces to a tab

        Argument:
            line: line to count tabs in

        Returns integer with number of tabs
    """
    for i in range(len(line)):
        if line[i] != '':
            return i
    return len(line)

def print_struct_outline(data_structure : Dict, indent_level : int = 0):
    """
        Prints structure outline to command line to provide user output

        Arguments:
            data_structure: data structure to print
            indent_level: current level of indent for recursive purposes
    """
    for key in data_structure:
        line_string = build_indent_string_print(indent_level)
        line_string += key
        print(line_string)
        if data_structure[key][0]:
            print_struct_outline(data_structure[key][0], indent_level + 1)



def build_indent_string_print(indent_level : int):
    """
        Builds a visibile indent indication for printing to command line

        Arguments:
            indent_level: level of indentation for the symbol
    """
    result = ''
    if indent_level == 0:
        return result
    indent_level -= 1
    for _ in range(indent_level):
        result += '\t'
    result += '|___'
    return result


from pyzbar.pyzbar import decode, ZBarSymbol
import cv2
from typing import List, Tuple, Dict
from multiprocessing import Pool, cpu_count
import os
import shutil
from functools import partial
import tqdm
import imghdr
import re

def read_qr_zbar(image_path : str, binarization : bool = False) -> List:
    """
        Loads and scans image for qr code using pyzbar. Processes raw image first. If it fails to find a QR code will
        try again with binarization using OpenCV with OTSU threshold finding.

        Arguments:
            image_path: path to image to scan
        
        Returns pyzbar results in list
    """
    im = cv2.imread(image_path)
    result = decode(im, symbols = [ZBarSymbol.QRCODE])
    if not result and binarization:
        im_gray = cv2.cvtColor(im, cv2.COLOR_BGR2GRAY)
        thres, im_binarized = cv2.threshold(im_gray, 128, 255, cv2.THRESH_OTSU)
        result = decode(im_binarized, symbols = [ZBarSymbol.QRCODE])
    return result

def get_qr(image_path : str, string_header : str = '', binarization : bool = False) -> str:
    """
        Loads image and attempts to find a QR code
    
        Arguments:
            image_path: string indicating path of photo to scan
            string_header: optional string that indicates the header to look  for if one was use
                if a header is passed only QR codes with this header will be returned

        Returns a string indicating content of found QR code. Returns None if nothing found.

        Does not currently handle multiple QR codes. Will raise exception
    """
    results = read_qr_zbar(image_path, binarization)
    valid_results : List(str) = []
    for result in results:
        str_data :str = result.data.decode('utf-8')
        if str_data.startswith(string_header):
            str_data_no_head = str_data[len(string_header):]
            valid_results.append(str_data_no_head)
    
    if len(valid_results) > 1:
        raise Exception('Found multiple valid QR codes. Could not conclusively pick path')
    elif len(valid_results) == 1:
        return valid_results[0]
    else:
        return None

def get_qr_mt(image_path : str, string_header : str = '', binarization : bool = False) -> Tuple[str, str]:
    """
        Calls get_qr and returns image path and result of get_qr in a tuple to allow multiprocessing

        Arguments:
            image_path: string indicating path of photo to scan
            string_header: optional string that indicates the header to look  for if one was use
                if a header is passed only QR codes with this header will be returned

    """
    return image_path, get_qr(image_path, string_header, binarization)

def get_qr_for_files(files : List[str], string_header : str = '', verbose : bool = False, binarization : bool = False) -> Dict[str, str]:
    '''
        Process a list of files using multiprocessing. Will return dictionary of each images'
        result against it's filename as a key

        Arguments:
            files: list of file paths as string
            string_header: string header used in images if any
            verbose: boolean indicating whether to show progress updates to terminal

        Returns:
            Dictionary of (str, str) where key is file path and value is the result
    '''
    cores = cpu_count()
    func = partial(get_qr_mt, string_header=string_header, binarization=binarization)
    with Pool(processes=cores) as pool:
        if verbose:
            results = list(tqdm.tqdm(pool.imap(func, files), total=len(files)))
        else:
            results = pool.map(func, files)
    
    results_dict = {}
    for item in results:
        results_dict[item[0]] = item[1]

    return results_dict

def sanitise_path(path : str) -> str:
    '''
        Function to remove invalid path characters from a path. Will prevent failure if user uses
        characters such as question marks in image strings.

        Arguments:
            path: path to sanitise

        Returns:
            sanitised path
    '''
    drive_designator_pos = path.find(':\\') + 2
    if drive_designator_pos > 0: #For windows with :\ in the tart of the path
        windows_drive_designator = path[:drive_designator_pos]
        remainder_path = path[drive_designator_pos:]
    else: #If windows drive designator not present (e.g. on Linux or with relative pathing)
        windows_drive_designator = ''
        remainder_path = path
    remainder_path = re.sub(r'[^\w\-_\. \\\/]', '_', remainder_path)
    return windows_drive_designator + remainder_path

def check_if_image(file_path : str) -> Tuple[bool, str]:
    '''
        Function to check if a file is an image. If the file is an image it will return True and the filepath,
        if it is not an image it will return False and the filepath.

        Arguments:
            file_path: path to item in question

        Returns:
            tuple of boolean and string indicating whether it is am image or not and the filepath of the associated check
    '''
    try:
        if imghdr.what(file_path):
            return (True, file_path)
    except:
        return (False, file_path)
    return (False, file_path)

def remove_non_images(files : List[str], verbose : bool, non_image_dir : str) -> List[str]:
    '''
        Function to remove items from the file list if they are not images. Non images files will be
        copied to the non-image directory so it is clear to the user what has happened with them.

        Parameters:
            files: list of files to check for image-ness
            verbose: indicates whether to provide verbose output to the user
            non_image_dir: directory to copy non-image files to for use feedback

        Returns:
            List of all files which are images
    '''
    if verbose:
        print("Checking for non-image files in the sorting directory. If any exist these will be saved to: " + non_image_dir)
    cores = cpu_count()
    with Pool(processes=cores) as pool:
        if verbose:
            results = list(tqdm.tqdm(pool.imap(check_if_image, files), total=len(files)))
        else:
            results = pool.map(check_if_image, files)

    new_files = []
    for is_image, file_path in results:
        if is_image:
            new_files.append(file_path)
        elif not os.path.isdir(file_path):
            os.makedirs(non_image_dir, exist_ok=True)
            _, file = os.path.split(file_path)
            shutil.copyfile(file_path, os.path.join(non_image_dir, file))
    
    return new_files



def get_image_paths(input_dir : str,  non_image_dir : str, verbose : bool = False) -> List[str]:
    """
        Gets all images in the provided input directory. Will exclude non-image files.

        Parameters:
            input_dir: Path to input directory containing photos as a string
            non_image_dir: Path to copy non-image files to
            verbose: Boolean indicating whether to print status information to command line

        Returns:
            List[str] of all iamge paths
    """
    images = os.listdir(input_dir)
    image_paths = [os.path.join(input_dir, x) for x in images]
    
    image_paths = remove_non_images(image_paths, verbose, non_image_dir)
    return image_paths


def sort_directory(input_dir : str, output_dir : str, string_header : str = '', verbose : bool = False, binarization : bool = False) -> List[str]:
    """
        Takes all images in a directory and sorts them by QR codes found in the images. Any
        images which are found before the first QR code will go into an "unsorted" folder in the directory.

        Parameters:
            input_dir: input directory containing photos as a string
            output_dir: target directory for photos, will be created if does not exist
            string_header: if a header is used in the QR codes to differentiate from other QR
                codes in the images, QR codes will be checked to ensure that the strings start
                with this substring

        Returns:
            List[str] of all paths found in QR codes
    """

    

    non_image_dir = os.path.join(output_dir, 'non_image_files')

    image_paths = get_image_paths(input_dir, non_image_dir, verbose)

    if verbose:
        print('Scanning images for QR codes')
    results = get_qr_for_files(image_paths, string_header=string_header, verbose=verbose, binarization=binarization)
    os.makedirs(output_dir, exist_ok=True)


    found_directories = sort_directory_exisitng_results(results, input_dir, output_dir, verbose)

    found_directories.sort()
    return found_directories

def sort_directory_exisitng_results(results : Dict[str, str], input_dir : str, output_dir : str, verbose : bool = False) -> List[str]:
    """
        Takes results from the QR code scanning and uses that information to sort the images. Function separated from
        above for better integration with GUI code.

        Parameters:
            results: Dictionary of results including image path and QR code result
            input_dir: input directory containing photos as string
            output_dir: output directory to save images in
            verbose: whether or not to write verbose output to the terminal

        Returns:
            List[str] of all paths found in QR codes
    """
    found_directories = []
    non_image_dir = os.path.join(output_dir, 'non_image_files')
    image_paths = get_image_paths(input_dir, non_image_dir, verbose)

    if verbose:
        print('Sorting image files')
    current_path = os.path.join(output_dir, 'unsorted')
    for image_path in tqdm.tqdm(image_paths) if verbose else image_paths:
        _, image = os.path.split(image_path)
        qr_string = results[image_path]
        if qr_string:
            qr_string = sanitise_path(qr_string)
            current_path = os.path.join(output_dir, qr_string)
            if qr_string not in found_directories:
                found_directories.append(qr_string)
            
        os.makedirs(current_path, exist_ok=True)
        shutil.copyfile(image_path, os.path.join(current_path, image))
    
    return found_directories



from os import listdir
from qrImageIndexer import photo_sorter, qr_generator
import pathlib
import os
from typing import Dict

def generate_image(tmp_path : pathlib.Path, string : str, filename : str = 'test') -> str:
    '''
        Generates a temporary image and returns the path to the image
        as a string.

        Arguments:
            tmp_path : pathlib.Path: Temporary path in which to create the image
            string : str: string to embed in qr code

        Returns path to image as str
    '''
    image = qr_generator.build_qr(string)
    image.save(tmp_path / (filename + '.png'))
    return (tmp_path / (filename + '.png')).as_posix()

def test_path_sanitise_already_clean():
    test_path_clean_windows = r'C:\Test_Path\No Invalid Characters\Test.File'
    sanitised = photo_sorter.sanitise_path(test_path_clean_windows)
    assert sanitised == test_path_clean_windows

    test_path_clean_unix = r'/usr/test/test_path/all valid chars.extn'
    sanitised = photo_sorter.sanitise_path(test_path_clean_unix)
    assert sanitised == test_path_clean_unix

def test_path_sanitise_unlcean():
    test_path_windows = r'C:\Test_Path\Bad:Chars\Test?'
    expected_path_windows = r'C:\Test_Path\Bad_Chars\Test_'
    assert photo_sorter.sanitise_path(test_path_windows) == expected_path_windows

    test_path_unix = r'/usr/test_path/test?invalid/Chars"'
    expected_path_unix = r'/usr/test_path/test_invalid/Chars_'
    assert photo_sorter.sanitise_path(test_path_unix) == expected_path_unix

''' 
    Acknowledge that this uses the image generation as an input and image generation uses this as a test,
    but it is vanishingly unlikely that zbar reads the QR code in the exactly incorrect way that the qr
    library generates it. Also assuming external libraries are OK.
'''
def test_read_qr_zbar(tmp_path : pathlib.Path):
    assert photo_sorter.read_qr_zbar(generate_image(tmp_path, 'test_string'))[0].data.decode('utf-8') == 'test_string'
    assert photo_sorter.read_qr_zbar(generate_image(tmp_path, 'test_string'))[0].data.decode('utf-8') == photo_sorter.read_qr_zbar(generate_image(tmp_path, 'test_string'), binarization=True)[0].data.decode('utf-8')

def test_get_qr_no_header(tmp_path : pathlib.Path):
    assert photo_sorter.get_qr(generate_image(tmp_path, 'test_string')) == 'test_string'

def test_get_qr_header(tmp_path : pathlib.Path):
    assert photo_sorter.get_qr(generate_image(tmp_path, r'{header}test_string'), r'{header}') == 'test_string'
    assert photo_sorter.get_qr(generate_image(tmp_path, r'{wrong_header}test_string'), r'{header}') == None
    assert photo_sorter.get_qr(generate_image(tmp_path, r'test_string'), r'{header}') == None

def test_get_qr_mt(tmp_path : pathlib.Path):
    path = generate_image(tmp_path, r'{header}test_string')
    assert photo_sorter.get_qr_mt(generate_image(tmp_path, r'{header}test_string'), r'{header}') == (path, 'test_string')
    assert photo_sorter.get_qr_mt(generate_image(tmp_path, r'{wrong_header}test_string'), r'{header}') == (path, None)
    assert photo_sorter.get_qr_mt(generate_image(tmp_path, r'test_string'), r'{header}') == (path, None)

def test_get_qr_for_files(tmp_path : pathlib.Path):
    files = []
    for i in range(5):
        content = 'content' + str(i)
        files.append(generate_image(tmp_path, content, str(i)))
    
    result : Dict[str, str] = photo_sorter.get_qr_for_files(files)
    for i in range(5):
        assert files[i] in result
        content = 'content' + str(i)
        assert result[files[i]] == content

    for i in range(5, 7):
        content = r'{content}' + str(i)
        files.append(generate_image(tmp_path, content, str(i)))

    result : Dict[str, str] = photo_sorter.get_qr_for_files(files, string_header=r'{content}')

    for i in range(5):
        assert files[i] in result
        assert result[files[i]] is None

    for i in range(5, 7):
        assert files[i] in result
        content = str(i)
        assert result[files[i]] == content


def test_is_image_true(tmp_path : pathlib.Path):
    image = qr_generator.build_qr('test')
    image.save(tmp_path / ('test.png'))
    
    assert photo_sorter.check_if_image((tmp_path / ('test.png')).as_posix())[0]

def test_is_image_false(tmp_path : pathlib.Path):
    temp_file = tmp_path /'test'
    temp_file.touch()
    
    assert not photo_sorter.check_if_image(temp_file.as_posix())[0] 

def test_remove_non_images(tmp_path : pathlib.Path):
    files = []
    
    non_image_dir = tmp_path/'non_image'
    non_image_dir.mkdir()

    im_path = generate_image(tmp_path, 'test')
    files.append(im_path)
    non_image_file = tmp_path/'non_image_file'
    non_image_file.touch()
    files.append(non_image_file.as_posix())

    results = photo_sorter.remove_non_images(files, False, non_image_dir.as_posix())

    assert len(results) == 1
    assert results[0] == im_path
    assert os.listdir(non_image_dir) == ['non_image_file']

def test_get_images(tmp_path : pathlib.Path):    
    non_image_dir = tmp_path/'non_image'
    non_image_dir.mkdir()

    im_path = generate_image(tmp_path, 'test')
    non_image_file = tmp_path/'non_image_file'
    non_image_file.touch()

    results = photo_sorter.get_image_paths(tmp_path.as_posix(), non_image_dir.as_posix())

    assert len(results) == 1
    assert results[0].replace('\\','/') == im_path # Since using os.join this will end up with Windows pathing in Windows - still OK
    assert os.listdir(non_image_dir) == ['non_image_file']

def test_qr_sorting(tmp_path : pathlib.Path):
    qr_strings = [r'Test1\subTest', r'Test1', r'Test2', r'Test?']
    inputs = tmp_path/'inputs'
    outputs = tmp_path/'outputs'
    outputs.mkdir()
    inputs.mkdir()
    for i in range(len(qr_strings)):
        image = qr_generator.build_qr(qr_strings[i])
        image.save(inputs / (str(i) + '.png'))


    found_dirs = photo_sorter.sort_directory(inputs.as_posix(), outputs.as_posix())

    assert found_dirs == [r'Test1', r'Test1\subTest', r'Test2', r'Test_']

    for i in range(len(qr_strings)):
        path = outputs/photo_sorter.sanitise_path(qr_strings[i])
        filecount = 0
        for filename in os.listdir(path.as_posix()):
            if os.path.isfile(path/filename): filecount += 1
        assert filecount == 1
        assert os.listdir(path.as_posix())[0] == (str(i) + '.png')



from fpdf import FPDF
from typing import Dict, List, Tuple
from PIL.Image import Image
from qrcode.image.pil import PilImage
from math import ceil
from copy import copy
import io

TARGET_ROWS_PER_PAGE = 6

def build_pdf_report(data_struct : Dict[str, Tuple[Dict, Image]], repeat_headings : bool = False, order_for_slicing : bool = False) -> FPDF:
    '''
        Builds a PDF report using FPDF2

        Arguments:
        data_struct : Dict[str, Tuple[Dict, Image]]: Input recursive data structure
        path: str: save path for PDF
        repeat_headings: bool: Indicates whether headings to be reapeated on each row or only at top level
        order_for_slicing: bool: whether to reorder table for slicing or have it down the page

    '''
    pdf = FPDF(orientation='portrait', format='A4')
    pdf.set_font('helvetica')
    pdf.add_page()
    data_table = build_data_table(data_struct, repeat_headings)
    if order_for_slicing:
        data_table = sort_table_for_slicing(data_table, TARGET_ROWS_PER_PAGE)
    
    cols = len(data_table[0])
    col_width = pdf.epw / cols
    line_height = (pdf.eph / (TARGET_ROWS_PER_PAGE)) * 0.99 #Going exact will cause a cautious page break

    for row in data_table:
        first_elem = True
        for datum in row:
            if first_elem:
                pdf.set_font(style='B')
                first_elem = False
            else:
                pdf.set_font(style='')
            if type(datum) is PilImage:
                f = io.BytesIO()
                temp : PilImage = datum
                temp.save(f)
                pdf.image(f, x=pdf.get_x(), y=pdf.get_y(), w=min(col_width, line_height))
                pdf.multi_cell(col_width, line_height, '', border = 1,
                    new_x="RIGHT", new_y="TOP", max_line_height=pdf.font_size)
            else:
                pdf.multi_cell(col_width, line_height, datum, border = 1,
                    new_x="RIGHT", new_y="TOP", max_line_height=pdf.font_size)
        pdf.ln(line_height)
    return pdf

def build_data_table(data_struct: Dict[str, Tuple[Dict, Image]], pass_headings : bool = False, existing_data : List = []) -> List[List]:
    '''
        Function to recursively turn the data structure into a nested list. Will include headings if required

        Arguments:
        data_struct: Dict[str, Tuple[Dict, Image]]: Nested data structure containing headings and QR images as appropriate.
        pass_headings: bool: Indicates whether headings to be included at each subsequent level
        existing_data: List: For recursive calls. Includes either heading data or empty strings depending on status of pass_headings argument

        Returns:
        List[List]: Containing data to be written to PDF table
    '''
    result = []

    for key in data_struct:
        row = copy(existing_data)
        result.append(row)
        row.append(key)

        if data_struct[key][0]:
            passed_row = copy(row)
            if not pass_headings:
                passed_row = ['' for _ in passed_row]
            for sub_row in build_data_table(data_struct[key][0], pass_headings, passed_row):
                result.append(sub_row)

        if data_struct[key][1]:
            row.append(data_struct[key][1])
        else:
            row.append('')

    max_cols = max([len(r) for r in result])
    for row in result:
        while len(row) < max_cols:
            row.insert(-1, '') # Pad out columns before QR for shorter rows so QRs are always in the last cell and each row is equal
    
    return result

def sort_table_for_slicing(input_table : List[List], rows_per_page : int) -> List[List]:
    '''
        Function to sort table so that it will be in an order that will allow
        easy sorting of QR code strips through the page. I.e. at the top of each
        page will be QR code 1, 2, 3, 4 etc. instead of down the page, such that when
        sliced into strips they will be in order when stacked.

        Arguments:
        input_table: List[List]: Table to be sorted. Expected that all rows same length and of at least length 1
        rows_per_page: int: Number of rows to be sorted for each page

        Returns
        List[List]: Sorted list of same number of elements as input sorted down the page
    '''    
    total_rows = len(input_table)
    page_count = ceil(total_rows/rows_per_page) # Need an extra full page if any remainder
    output = []
    for page in range(page_count):
        for line in range(rows_per_page):
            index = page + line * page_count
            if index < total_rows:
                output.append(input_table[index])
            else:
                output.append(['' for _ in input_table[0]])
    
    return output
    




from qrImageIndexer import generate_qr_wrapper
from unittest import mock

def mock_qr_generator_functions(mocker : mock):
    unpack_data = mocker.patch('qrImageIndexer.generate_qr_wrapper.unpack_data')
    generate_qr_structure = mocker.patch('qrImageIndexer.generate_qr_wrapper.generate_qr_code_structure')
    build_pdf = mocker.patch('qrImageIndexer.generate_qr_wrapper.build_pdf_report')

    return (unpack_data, generate_qr_structure, build_pdf)


def test_single_func_pdf_builder_all_true(mocker):
    (unpack_data, generate_qr_structure, build_pdf) = mock_qr_generator_functions(mocker)
    result = generate_qr_wrapper.generate_qr_pdf([['Line 1', ''], ['', 'Line 1 Indent']],
                                        True, True, True, r'{header}')
    
    unpack_data.assert_called_once_with([['Line 1', ''], ['', 'Line 1 Indent']], True, r'{header}')
    generate_qr_structure.assert_called_once_with(unpack_data.return_value)
    build_pdf.assert_called_once_with(generate_qr_structure.return_value, True, True)
    assert result == build_pdf.return_value

def test_single_func_pdf_builder_no_qr_heading(mocker):
    (unpack_data, generate_qr_structure, build_pdf) = mock_qr_generator_functions(mocker)
    result = generate_qr_wrapper.generate_qr_pdf([['Line 1', ''], ['', 'Line 1 Indent']],
                                        False, True, True, r'{header}')
    
    unpack_data.assert_called_once_with([['Line 1', ''], ['', 'Line 1 Indent']], False, r'{header}')
    generate_qr_structure.assert_called_once_with(unpack_data.return_value)
    build_pdf.assert_called_once_with(generate_qr_structure.return_value, True, True)
    assert result == build_pdf.return_value

def test_single_func_pdf_builder_no_title_all_heading(mocker):
    (unpack_data, generate_qr_structure, build_pdf) = mock_qr_generator_functions(mocker)
    result = generate_qr_wrapper.generate_qr_pdf([['Line 1', ''], ['', 'Line 1 Indent']],
                                        True, False, True, r'{header}')
    
    unpack_data.assert_called_once_with([['Line 1', ''], ['', 'Line 1 Indent']], True, r'{header}')
    generate_qr_structure.assert_called_once_with(unpack_data.return_value)
    build_pdf.assert_called_once_with(generate_qr_structure.return_value, False, True)
    assert result == build_pdf.return_value

def test_single_func_pdf_builder_not_sliceable(mocker):
    (unpack_data, generate_qr_structure, build_pdf) = mock_qr_generator_functions(mocker)
    result = generate_qr_wrapper.generate_qr_pdf([['Line 1', ''], ['', 'Line 1 Indent']],
                                        True, True, False, r'{header}')
    
    unpack_data.assert_called_once_with([['Line 1', ''], ['', 'Line 1 Indent']], True, r'{header}')
    generate_qr_structure.assert_called_once_with(unpack_data.return_value)
    build_pdf.assert_called_once_with(generate_qr_structure.return_value, True, False)
    assert result == build_pdf.return_value


from distutils.command import build
from typing import List
from fpdf import FPDF
from qrImageIndexer.qr_generator import unpack_data, generate_qr_code_structure
from qrImageIndexer.write_pdf_fpf2 import build_pdf_report

def generate_qr_pdf(text_data: List[List[str]], qr_for_headings : bool,
                    repeat_headings : bool, sort_sliceable : bool, string_header : str = '') -> FPDF:
    data_struct = unpack_data(text_data, qr_for_headings, string_header)
    iamge_struct = generate_qr_code_structure(data_struct)
    return build_pdf_report(iamge_struct, repeat_headings, sort_sliceable)


from qrImageIndexer.qr_generator import load_text_file, print_struct_outline, unpack_data, generate_qr_code_structure
from qrImageIndexer.generate_qr_wrapper import generate_qr_pdf
from qrImageIndexer.photo_sorter import sort_directory
from qrImageIndexer.write_pdf_fpf2 import build_pdf_report

import argparse

def main():
    parser = argparse.ArgumentParser()
    mutual_exclusive = parser.add_mutually_exclusive_group()
    mutual_exclusive.add_argument('-g', '--generate-pdf', help='Generate a PDF of QR codes from a given text file. Specify ',
            nargs=2, metavar=('INPUT_TEXT_FILE','OUTPUT_PDF'))
    mutual_exclusive.add_argument('-s', '--sort-photos', help='Sort photos based on QR codes found in photos. Once a QR code is found all photos will be sorted into the directory indicated by the code until subsequent codes found',
            nargs=2, metavar=('INPUT_DIR', 'OUTPUT_DIR'))
    parser.add_argument('--pdf-type', help='Type of PDF to generate. Either linearly sorted or sorted to enable easy slicing of the printed pages. Accepts "linear" or "sliceable". Linear will sort down page, sliceable will sort "through" the page.',
            nargs=1, metavar=('SORT_TYPE'), choices=['linear', 'sliceable'])

    parser.add_argument('-q', '--qr-for-headings', help='Generate a QR code for each heading, not just a code for the last items in a tree.',
            action='store_true')
    parser.add_argument('-r', '--repeat-table-headings', help='Repeat table headings on every line',
            action='store_true')

    # parser.add_argument('-b', '--use-binarization', help='Search for QR codes with binarized images. May take up to twice as long but can sometimes find QR codes that would otherwise not be detected.',
    #     action='store_true')

    parser.add_argument('-p', '--string-prefix', help='Specify a prefix for use in the generated QR codes to differentiate from codes that might also end up in photos')

    parser.add_argument('-v', '--verbose', help='Turn progress text to terminal on or off',
            action='store_true')

    args = parser.parse_args()

    string_header = ''
    if args.string_prefix:
        string_header = args.string_prefix

    verbose = args.verbose
    
    if not args.generate_pdf and not args.sort_photos:
        print("Neither sort photos or generate PDF was selected. No action performed.")
    if args.generate_pdf:
        input = args.generate_pdf[0]
        output = args.generate_pdf[1]

        sliceable = False
        if args.pdf_type:
            sliceable = args.pdf_type[0] == 'sliceable'

        text_data = load_text_file(input)
        
        if verbose:
            print('Loaded text file: ' + input)
            print('')
            print('Read lines: ')
            print(text_data)

        pdf = generate_qr_pdf(text_data, args.qr_for_headings, args.repeat_table_headings, sliceable, string_header)
        pdf.output(output)
        if verbose:
            print('Saved pdf: ' + output)
    if args.sort_photos:
        input = args.sort_photos[0]
        output = args.sort_photos[1]

        if verbose:
            print('Sorting from: ' + input + ', to : ' + output)

        found_dirs = sort_directory(input, output, string_header, args.verbose, True)

        if verbose:
            print('Found directories from images:')
            [print(x) for x in found_dirs]
        
        


if __name__ == '__main__':
    main()


from qrImageIndexer import write_pdf_fpf2
from .test_qr_generator import demo_data_struct_include_headers
from os import path
import pathlib
from PyPDF2 import PdfReader

def expected_table_include_headers_strings_repeat_headings():
    expected_structure = [
        ['Level1-1', '', '', 'Level1-1' + path.sep],
        ['Level1-2', '', '', 'Level1-2' + path.sep],
        ['Level1-2', 'Level2-1', '', 'Level1-2' + path.sep + 'Level2-1' + path.sep],
        ['Level1-2', 'Level2-2', '', 'Level1-2' + path.sep + 'Level2-2' + path.sep],
        ['Level1-2', 'Level2-2', 'Level3', 'Level1-2' + path.sep + 'Level2-2' + path.sep + 'Level3' + path.sep]
    ]
    return expected_structure

def expected_table_include_headers_strings_no_repeat_headings():
    expected_structure = [
        ['Level1-1', '', '', 'Level1-1' + path.sep],
        ['Level1-2', '', '', 'Level1-2' + path.sep],
        ['', 'Level2-1', '', 'Level1-2' + path.sep + 'Level2-1' + path.sep],
        ['', 'Level2-2', '', 'Level1-2' + path.sep + 'Level2-2' + path.sep],
        ['', '', 'Level3', 'Level1-2' + path.sep + 'Level2-2' + path.sep + 'Level3' + path.sep]
    ]
    return expected_structure

def test_build_table_repeat_headings():
    assert expected_table_include_headers_strings_repeat_headings() == write_pdf_fpf2.build_data_table(demo_data_struct_include_headers(), True)

def test_build_table_no_repeat_headings():
    assert expected_table_include_headers_strings_no_repeat_headings() == write_pdf_fpf2.build_data_table(demo_data_struct_include_headers(), False)

def test_sort_table_for_slicing_no_remainder():
    input = [
        [1],
        [2],
        [3],
        [4],
        [5],
        [6],
        [7],
        [8]
    ]

    expected = [
        [1],
        [3],
        [5],
        [7],
        [2],
        [4],
        [6],
        [8]
    ]

    assert expected == write_pdf_fpf2.sort_table_for_slicing(input, 4)

def test_sort_table_for_slicing_remainder():
    input = [
        [1],
        [2],
        [3],
        [4],
        [5],
        [6],
        [7],
        [8]
    ]

    expected = [
        [1],
        [4],
        [7],
        [2],
        [5],
        [8],
        [3],
        [6],
        ['']
    ]

    assert expected == write_pdf_fpf2.sort_table_for_slicing(input, 3)

def test_generated_pdf_text(tmp_path : pathlib.Path):
    data_dict = {
        'Line 1':
            ({'Indent 1' : 
                ({'Indent 2' : ({}, None)},
                    None)},
                None),
        'Line 2': 
            ({}, None)
    }
    path = tmp_path / 'test.pdf'
    pdf = write_pdf_fpf2.build_pdf_report(data_dict)
    pdf.output(path.as_posix())
    reader = PdfReader(path)
    page = reader.pages[0]
    text = page.extract_text()
    assert 'Line 1' in text
    assert 'Line 2' in text
    assert 'Indent 1' in text
    assert 'Indent 2' in text

    